/*
** wifimgr.c
**	manage WiFi networks
**	scans for available WiFi networks
**	edits /etc/wpa_supplicant.conf
*/

/*
** Copyright (c) 2009, J.R. Oldroyd, Open Advisors Limited
** All rights reserved.
** 
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the author, the author's organization nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
** 
** THIS SOFTWARE IS PROVIDED BY OPEN ADVISORS LIMITED ''AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL OPEN ADVISORS LIMITED BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
** $Id: wifimgr.c 57 2009-12-18 19:29:30Z jr $
*/

#include "wifimgr.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#ifndef WITHOUT_NLS
#include <libintl.h>
#else
#define gettext(x)	(x)
#endif

char *			wifi_if;
struct wifi_net *	nets;
FILE *			sucmd;

/*
** parse /etc/rc.conf to find WiFi interface
*/
char *
find_wifi_if(char * file) {
	FILE *			fp;
	char			line[1024];
	char *			p;
	static char		wifi_if[256];

	if ((fp = fopen(file, "r")) == NULL) {
		char		buf[256];
		sprintf(buf, gettext("Cannot open <b>%s</b>."), file);
		gui_message(buf, MSG_ERROR);
		exit(1);
	}

	while(fscanf(fp, "%[^\n]\n", line) != EOF) {
		for (p = line; *p == ' ' || *p == '\t'; p++);
		if (*p == '#')
			continue;
		if (strncasecmp(p, "ifconfig_", 9) == 0
		    && strcasestr(p, "=\"")
		    && strcasestr(p, "WPA")
		    && strcasestr(p, "DHCP")
		) {
			sscanf(p, "ifconfig_%[^=]=", wifi_if);
			return wifi_if;
		}
	}

	return NULL;
}

/*
** parse /etc/wpa_supplicant.conf for currently configured networks
*/
void
read_networks_file(char * file) {
	char			line[1024];
	char			resp[256];
	char *			p;
	struct wifi_net *	new;
	int			extra_config = 0;

	fprintf(sucmd, "cat_netfile\n");

	while(fscanf(sucmd, "%[^\n]\n", line) != EOF) {
		if (strcmp(line, "EOF") == 0)
			break;
		for (p = line; *p == ' ' || *p == '\t'; p++);
		if (strcmp(p, "network={") == 0) {
			if ((new = (struct wifi_net *) malloc(sizeof(struct wifi_net))) == NULL) {
				char		buf[256];
				sprintf(buf, gettext("Out of memory."));
				gui_message(buf, MSG_ERROR);
				exit(1);
			}
			new->wn_enabled = 1;
			*new->wn_ssid = '\0';
			*new->wn_key = '\0';
			*new->wn_comment = '\0';
			new->wn_priority = 0;
			*new->wn_bssid = '\0';
			new->wn_chan = 0;
			new->wn_rate = 0;
			new->wn_bars = 0;
			new->wn_security = WN_SEC_NONE;
			new->wn_next = NULL;
		}
		else if (strncmp(p, "#: ", 3) == 0) {
			for (p += 3; *p == ' ' || *p == '\t'; p++);
			strcpy(new->wn_comment, p);
		}
		else if (*p == '#')
			continue;
		else if (strncmp(p, "ssid=", 5) == 0) {
			sscanf(p, "ssid=\"%[^\"]\"", new->wn_ssid);
			if (*new->wn_ssid == '*') {
				strcpy(new->wn_ssid, &new->wn_ssid[1]);
				new->wn_priority = -1;
			}
		}
		else if (strncmp(p, "bssid=", 6) == 0)
			strcpy(new->wn_bssid, p+6);
		else if (strncmp(p, "priority=", 9) == 0)
			sscanf(p, "priority=%d", &new->wn_priority);
		else if (strcmp(p, "key_mgmt=NONE") == 0) {
			new->wn_security = WN_SEC_NONE;
			continue;
		}
		else if (strcmp(p, "key_mgmt=WPA-PSK") == 0) {
			new->wn_security = WN_SEC_WPA_PSK;
			continue;
		}
		else if (strcmp(p, "key_mgmt=WPA-EAP") == 0 || strcmp(p, "key_mgmt=IEEE8021X") == 0) {
			new->wn_security = WN_SEC_OTHER;
			continue;
		}
		else if (strncmp(p, "wep_tx_keyidx=", 14) == 0)
			continue;
		else if (strncmp(p, "psk=", 4) == 0) {
			sscanf(p, "psk=\"%[^\"]\"", new->wn_key);
			new->wn_security = WN_SEC_WPA_PSK;
		}
		else if (strncmp(p, "wep_key0=", 9) == 0) {
			if (index(p, '"'))
				sscanf(p, "wep_key0=\"%[^\"]\"", new->wn_key);
			else
				sscanf(p, "wep_key0=%[^\n]\n", new->wn_key);
			new->wn_security = WN_SEC_WEP;
		}
		else if (*p == '}')
			nl_insert(&nets, new);
		else {
			extra_config++;
		}
	}
	fscanf(sucmd, "%[^\n]%*c", resp);
	if (strcmp(resp, "OK") != 0) {
		char		buf[256];
		char *		err;

		err = index(resp, ' ') + 1;
		sprintf(buf, gettext("Cannot read configuration file <b>%s</b> - %s."), file, err);
		gui_message(buf, MSG_ERROR);
		exit(1);
	}

	/* if unknown config directives were found, try to make backup copy */
	if (extra_config) {
		char		resp[256];
		char		save_file[256];

		sprintf(save_file, "%s.save", file);
		if (eaccess(save_file, F_OK) == 0) {
			fprintf(sucmd, "diff_netfile\n");
			fscanf(sucmd, "%[^\n]%*c", resp);
			if (strcmp(resp, "OK") != 0) {
				char		buf[256];
				/* NOTE TO TRANSLATOR: leave "%s.save" in English */
				sprintf(buf, gettext("Configuration file contains additional directives.\n"
						     "A backup file <b>%s.save</b> with\n"
						     "different contents already exists.  You must rename\n"
						     "the configuration file by hand."), file);
				gui_message(buf, MSG_ERROR);
				exit(1);
			}
		}
		fprintf(sucmd, "backup_netfile\n");
		fscanf(sucmd, "%[^\n]%*c", resp);
		if (strcmp(resp, "OK") != 0) {
			char		buf[256];
			/* NOTE TO TRANSLATOR: leave "%s.save" in English */
			sprintf(buf, gettext("Cannot write backup configuration file <b>%s.save</b>."), file);
			gui_message(buf, MSG_ERROR);
			exit(1);
		}
		else {
			char		buf[256];
			/* NOTE TO TRANSLATOR: leave "%s.save" in English */
			sprintf(buf, gettext("Configuration file contains additional directives;\n"
				     "a backup copy has been made in <b>%s.save</b>."), file);
			gui_message(buf, MSG_INFO);
		}
	}
}

/*
** check for 10 or 26 char hex key
*/
int
wep_hexkey(char * s) {

	/* check length */
	if (!(strlen(s) == 10 || strlen(s) == 26))
		return 0;

	/* check for hex chars only */
	while (*s) {
		if (!((*s >= '0' && *s <= '9')
		 || (*s >= 'A' && *s <= 'F')
		 || (*s >= 'a' && *s <= 'f')))
			return 0;
		s++;
	}

	return 1;
}

/*
** (re-)write /etc/wpa_supplicant.conf
*/
int
save_networks_file() {
	char *			tmpdir;
	char			tmpfile[256];
	char			resp[256];
	FILE *			fp;
	struct wifi_net *	p;

	/* validate the key lengths */
	for (p = nets; p; p = p->wn_next)
		if (p->wn_enabled) {
			if (p->wn_security == WN_SEC_WPA_PSK) {
				if (strlen(p->wn_key) < 8) {
					char		buf[256];
					sprintf(buf, gettext("Invalid key for network <b>%s</b>.\n\n"
						     "WPA keys must be 8 characters or longer."), p->wn_ssid);
					gui_message(buf, MSG_WARNING);
					return 0;
				}
			}
			else if (p->wn_security == WN_SEC_WEP) {
				if (!(strlen(p->wn_key) == 5 || strlen(p->wn_key) == 13
				   || wep_hexkey(p->wn_key))) {
					char		buf[256];
					sprintf(buf, gettext("Invalid key for network <b>%s</b>.\n\n"
						     "WEP keys must be 5 or 13 character strings or\n"
						     "10 or 26 character hexadecimal numbers."), p->wn_ssid);
					gui_message(buf, MSG_WARNING);
					return 0;
				}
			}
		}

	umask(077);
	tmpdir = getenv("TMPDIR");
	if (!tmpdir)
		tmpdir = TMPDIR;
	sprintf(tmpfile, "%s/wfm.%06d", tmpdir, getpid());
	if ((fp = fopen(tmpfile, "w")) == NULL) {
		char		buf[256];
		sprintf(buf, gettext("Cannot write temporary networks file <b>%s</b>."), tmpfile);
		gui_message(buf, MSG_ERROR);
		exit(1);
	}

	fprintf(fp, "# /etc/wpa_supplicant.conf written by wifimgr(8)\n\n");

	for (p = nets; p; p = p->wn_next)
		if (p->wn_enabled) {
			fprintf(fp, "network={\n");
			if (p->wn_comment[0] != '\0')
				fprintf(fp, "	#: %s\n", p->wn_comment);
			if (p->wn_priority >= 0)
				fprintf(fp, "	ssid=\"%s\"\n", p->wn_ssid);
			else
				fprintf(fp, "	ssid=\"*%s\"\n", p->wn_ssid);
			if (p->wn_priority > 0)
				fprintf(fp, "	priority=%d\n", p->wn_priority);
			if (strcmp(p->wn_bssid, "") != 0)
				fprintf(fp, "	bssid=%s\n", p->wn_bssid);
			if (p->wn_security == WN_SEC_WPA_PSK) {
				fprintf(fp, "	key_mgmt=WPA-PSK\n");
				fprintf(fp, "	psk=\"%s\"\n", p->wn_key);
			}
			else if (p->wn_security == WN_SEC_WEP) {
				fprintf(fp, "	key_mgmt=NONE\n");
				fprintf(fp, "	wep_tx_keyidx=0\n");
				if (!(strlen(p->wn_key) == 5 || strlen(p->wn_key) == 13))
					fprintf(fp, "	wep_key0=%s\n", p->wn_key);
				else
					fprintf(fp, "	wep_key0=\"%s\"\n", p->wn_key);
			}
			else if (p->wn_security == WN_SEC_NONE)
				fprintf(fp, "	key_mgmt=NONE\n");
			fprintf(fp, "}\n\n");
		}

	fclose(fp);

	/* execute su command to copy file */
	fprintf(sucmd, "write_netfile %s\n", tmpfile);
	fscanf(sucmd, "%[^\n]%*c", resp);
	unlink(tmpfile);
	if (strcmp(resp, "OK") != 0) {
		char		buf[256];
		char *		err;

		err = index(resp, ' ') + 1;
		sprintf(buf, gettext("Cannot write configuration file <b>%s</b> - %s."), NETWORKS_FILE, err);
		gui_message(buf, MSG_ERROR);
		exit(1);
	}

	/* restart interface */
	fprintf(sucmd, "restart_netif %s\n", wifi_if);
	fscanf(sucmd, "%[^\n]%*c", resp);
	if (strcmp(resp, "OK") != 0) {
		char		buf[256];
		char *		err;

		err = index(resp, ' ') + 1;
		sprintf(buf, gettext("Cannot reset interface <b>%s</b> - %s."), wifi_if, err);
		gui_message(buf, MSG_ERROR);
		exit(1);
	}

	/* clear changes flag */
	gui_changes = 0;

	return 1;
}

/*
** scan for currently available networks
*/
void
ifconfig_network_scan(char * intf) {
	char			cmd[256];
	FILE *			fp;
	char			ssid[256];
	char			bssid[256];
	int			chan;
	char			rate[256];
	int			sn_sig;
	int			sn_noise;
	char			caps[256];
	char			rest[1024];
	struct wifi_net *	new;

	sprintf(cmd, "%s -v %s list scan | %s 's/^ /-/'", PATH_IFCONFIG, intf, PATH_SED);

	if ((fp = popen(cmd, "r")) == NULL) {
		char		buf[256];
		sprintf(buf, gettext("Cannot exec %s."), PATH_IFCONFIG);
		gui_message(buf, MSG_ERROR);
		exit(1);
	}

	fscanf(fp, "%*[^\n]\n");	/* skip header line */
	while(fscanf(fp, "%32c %s %d %s %d:%d %*d %s %[^\n]\n",
	    ssid, bssid, &chan, rate, &sn_sig, &sn_noise, caps, rest) != EOF) {
		char *			p;
		for (p = &ssid[31]; p >= ssid && *p == ' '; p--)
			*p = '\0';
		if (*ssid == '-' && *(ssid+1) == '\0')
			sscanf(rest, "SSID<%[^>]>", ssid);

		if ((new = (struct wifi_net *) malloc(sizeof(struct wifi_net))) == NULL) {
			char		buf[256];
			sprintf(buf, gettext("Out of memory."));
			gui_message(buf, MSG_ERROR);
			exit(1);
		}

		new->wn_enabled = 0;
		strcpy(new->wn_ssid, ssid);
		*new->wn_key = '\0';
		*new->wn_comment = '\0';
		new->wn_priority = 0;
		strcpy(new->wn_bssid, bssid);
		new->wn_chan = chan;
		new->wn_rate = atoi(rate);
		new->wn_bars = (abs(sn_noise) - abs(sn_sig)) * 4;
		if (index(caps, 'P')) {
			if (strstr(rest, "WPA<")) {
				new->wn_security = WN_SEC_OTHER;
				if (strcasestr(rest, "8021x-PSK"))
					new->wn_security = WN_SEC_WPA_PSK;
			}
			else if (strstr(rest, "WPS<"))
				new->wn_security = WN_SEC_WPA_PSK;
			else
				new->wn_security = WN_SEC_WEP;
		}
		else
			new->wn_security = WN_SEC_NONE;
		new->wn_next = NULL;

		nl_insert(&nets, new);
	}

	pclose(fp);
}

/*
** run ifconfig to find active ssid
*/
char *
ifconfig_associated_network(char * intf) {
	char			cmd[256];
	FILE *			fp;
	char			line[1024];
	static char		ssid[256];

	/* run ifconfig command */
	sprintf(cmd, "%s %s", PATH_IFCONFIG, intf);
	if ((fp = popen(cmd, "r")) == NULL) {
		char		buf[256];
		sprintf(buf, gettext("Cannot exec %s."), PATH_IFCONFIG);
		gui_message(buf, MSG_ERROR);
		exit(1);
	}

	/* clear any previous ssid */
	ssid[0] = '\0';

	/* look for ssid */
	while(fscanf(fp, "%[^\n]\n", line) != EOF)
		if (sscanf(line, "ssid %s", ssid))
			break;

	/* close pipe */
	pclose(fp);

	return ssid;
}

/*
** build list of networks from existing config file and ifconfig scan
*/
struct wifi_net *
build_network_list(char * wifi_if) {

	if (nets)
		nl_delete_list(&nets);

	read_networks_file(NETWORKS_FILE);
	ifconfig_network_scan(wifi_if);

	return nets;
}

/*
** open setuid backend
*/
int
wifimgrsu_init()
{
	char *			pass;
	char			resp[256];

	if (gui_message(gettext("Enter Administrator password:"), MSG_INPUT) != MSG_RESPONSE_OK)
		exit(1);

	pass = gui_response;

	if ((sucmd = popen(PATH_WIFIMGRSU, "r+")) == NULL) {
		char		buf[256];
		sprintf(buf, gettext("Cannot exec %s."), PATH_IFCONFIG);
		gui_message(buf, MSG_ERROR);
		exit(1);
	}

	/* set line buffering */
	setvbuf(sucmd, (char *) NULL, _IOLBF, 0);

	fprintf(sucmd, "password %s\n", pass);
	fscanf(sucmd, "%[^\n]%*c", resp);
	if (strcmp(resp, "OK") != 0) {
		char		buf[256];
		char *		err;

		err = index(resp, ' ') + 1;
		sprintf(buf, gettext("Password error - %s."), err);
		gui_message(buf, MSG_ERROR);
		exit(1);
	}

	return 1;
}

int
main(int argc, char ** argv) {

#ifndef WITHOUT_NLS
	/* set output character set */
	bind_textdomain_codeset("wifimgr", "UTF-8");

	/* set language catalog domain */
	textdomain("wifimgr");
#endif

	/* open display */
	if (!gui_init(&argc, &argv)) {
		fprintf(stderr, gettext("wifimgr: cannot open display\n"));
		exit(1);
	}

	/* open channel to setuid backend */
	wifimgrsu_init();

	/* find WiFi interface */
	if ((wifi_if = find_wifi_if(RC_CONF_FILE)) == NULL) {
		char		buf[256];
		sprintf(buf, gettext("No WiFi interface is configured in <b>%s</b>."), RC_CONF_FILE);
		gui_message(buf, MSG_ERROR);
		exit(1);
	}

	gui_loop();

	exit(0);
}
